
/*		Copyright (c) 1987 Motorola Inc.		*/
/*		All Rights Reserved				*/
/*		@(#)z8530.m4	35.3	*/

#ident	"@(#)z8530.m4	35.3"

/**************************************************************
 *
 *	Driver for Z8530 Serial Communications Controller on
 *	the MVME134, MVME143 and the MVME147 boards.
 *
 *************************************************************/

/**********************	INCLUDES *****************************/

#include  "sys/param.h"
#include  "sys/types.h"
#include  "sys/dir.h"
#include  "sys/file.h"
#include  "sys/signal.h"
#include  "sys/tty.h"
#include  "sys/user.h"
#include  "sys/errno.h"
#include  "sys/termio.h"
#include  "sys/conf.h"
#include  "sys/sysinfo.h"
#include "sys/sysmacros.h"
#include "sys/gentty.h"
#include "sys/mvmecpu.h"

/*********************** CONFIGURATION *************************/

#define	MAXSCCS		2
#define MAXPORTS	(MAXSCCS * 2)

/**********************	CONSTANT DEFINITIONS *******************/

#define CR	0x0d
#define LF	0x0a	

#define PORT0	0
#define PORT1	1
#define PORT2	2
#define PORT3	3

/********************** COMPILER SWITCHES **********************/

#define	DEBUG 	 	/* turns on debug messages */
#undef DEBUG

/********************** Z8530 DEFINITIONS **********************/

/*	Write Register 0 - Command Register     */

#define	REG0		0x0		/* register selection codes */
#define REG1		0x1
#define REG2		0x2
#define REG3		0x3
#define REG4		0x4
#define REG5		0x5
#define REG6		0x6
#define REG7		0x7
#define REG8		0x8
#define REG9		0x9
#define REG10		0xA
#define REG11		0xB
#define REG12		0xC
#define REG13		0xD
#define REG14		0xE
#define REG15		0xF

#define RESINT		0x10		/* reset external/status interrupts */ 
#define SENDABT		0x18		/* send abort (SDLC) */
#define RXNXTEN		0x20		/* enable int on next Rx character */
#define RESTXIP		0x28		/* reset Tx int pending */
#define ERRRES		0x30		/* error reset */
#define RESHIUS		0x38		/* reset highest IUS */
#define RESRXCRC	0x40		/* reset Rx CRC checker */
#define RESTXCRC	0x80		/* reset Tx CRC generator */
#define RESTXURN	0xC0		/* reset Tx underrun */

/* 	Write Register 1 - Transmit/Receive Interrupt Register */

#define RXINTDIS	0x0		/* Rx int disable */
#define ESINTEN		0x1		/* external int enable */
#define TXINTEN		0x2		/* Tx int enable */
#define PARSRC		0x4		/* parity is special condition */
#define RXINT1ST	0x8		/* Rx int on first char or special 
					   receive condition (src) */
#define RXINTALL	0x10		/* int on all Rx chars or src */ 
#define RXINTSRC	0x18		/* Rx int on src only */
#define RXINTMSK	(RXINT1ST|RXINTALL|RXINTSRC)
#define WDORT		0x20		/* wait/DMA request on rec/trans */
#define WDRF		0x40		/* wait/DMA request function */
#define WDREN		0x80		/* wait/DMA request enable */
 
/* 	Write Register 2 - Interrupt Vector Register */

/*	Write Register 3 - Receive Parameters and Control */

#define RXENAB		0x1		/* receiver enable */
#define SYNCLIN		0x2		/* sync character load inhibit */
#define ADDSRCH		0x4		/* address search mode (SLDC) */
#define RXCRCEN		0x8		/* Rx CRC enable */
#define HUNTM		0x10		/* enter hunt mode */
#define AUTOEN		0x20		/* auto enables */
#define RX5BITS		0x0		/* Rx 5 bits/character */
#define RX7BITS		0x40		/* Rx 7 bits/character */
#define RX6BITS		0x80		/* Rx 6 bits/character */
#define RX8BITS		0xC0		/* Rx 8 bits/character */
#define RXCHMSK		(RX5BITS | RX6BITS | RX7BITS | RX8BITS)

/*	Write Register 4 - Transmit/Receive Miscellaneous Parameters */

#define ENPAR		0x1		/* parity enable */
#define EVENPAR		0x2		/* parity - 1 = even, 0 = odd */
#define SYNCEN		0x0		/* sync modes enable */
#define STBIT1		0x4		/* 1 stop bit/character */
#define STBIT1H		0x8		/* 1 1/2 stop bits/character */
#define STBIT2		0xC		/* 2 stop bits/character */
#define STBMSK		(STBIT1 | STBIT1H | STBIT2)
#define SY8BIT		0x0		/* 8 bit sync character */
#define SY16BIT		0x10		/* 16 bit sync character */
#define SDLCM		0x20		/* SDLC mode */
#define SYEXT		0x30		/* external sync mode */
#define X1CLK		0x0		/* x1 clock mode */
#define X16CLK		0x40		/* x16 clock mode */
#define X32CLK		0x80		/* x32 clock mode */
#define X64CLK		0xC0		/* x64 clock mode */

/*	Write Register 5 - Transmit Parameter and Control */

#define TXCRCEN		0x1		/* Tx CRC enable */
#define RTS		0x2		/* request to send */
#define SDLCRC16	0x4		/* SDLC/CRC-16 */
#define TXENAB		0x8		/* Tx enable */
#define SENDBR		0x10		/* send break */
#define TX5BITS		0x0		/* Tx 5 bits/character */
#define TX7BITS		0x20		/* Tx 7 bits/character */
#define TX6BITS		0x40		/* Tx 6 bits/character */
#define TX8BITS		0x60		/* Tx 8 bits/character */
#define TXCHMSK		(TX5BITS | TX6BITS | TX7BITS | TX8BITS)
#define DTR		0x80		/* data terminal ready */ 

/*	Write Register 9 - Master Interrupt Control */

#define VIS		0x1		/* vector includes status */
#define NV		0x2		/* no vector */
#define DLC		0x4		/* disable lower chain */
#define MIE		0x8		/* master interrupt enable */
#define SHSL		0x10		/* status high/status low */
#define CHRESB		0x40		/* channel reset B */
#define CHRESA		0x80		/* channel reset A */
#define FHRES		0xC0		/* force hardware reset */

/*	Write Register 11 - Clock Mode Control */

#define XTALOUT		0x0		/* TRxC OUT = XTAL output */
#define TXCOUT		0x1		/* TRxC OUT = transmit clock */
#define BRGOUT		0x2		/* TRxC OUT = BR generator output */
#define DPLLOUT		0x3		/* TRxC OUT = DPLL output */
#define TRXCOI		0x4		/* TRxC O/I */
#define TCRTXC		0x0		/* transmit clock = RTxC pin */
#define TCTRXC		0x8		/* transmit clock = TRxC pin */
#define TCBRGO		0x10		/* transmit clock = BRG output */
#define TCDPLLO		0x18		/* transmit clock = DPLL output */
#define RCRTXC		0x0		/* receive clock = RTxC pin */
#define RCTRXC		0x20		/* receive clock = TRxC pin */
#define RCBRGO		0x40		/* receive clock = BRG output */
#define RCDPLLO		0x60		/* receive clock = DPLL output */
#define RTXCXTAL	0x80		/* RTxC  XTAL/NO XTAL

/* 	Write Register 12 - Lower Byte of BRG Time Constant */

/*	Write Register 13 - Upper Byte of BRG Time Constant */

/*	Write Register 14 - Miscellaneous Control bits */

#define BRGEN 		0x1		/* BR generator enable */
#define BRGSRC		0x2		/* BR generator source */
#define DTRREQ		0x4		/* DTR/REQUEST function */
#define AUTOECHO	0x8		/* auto echo */
#define LCLLOOP		0x10		/* local loopback */
#define NULLCMD		0x0		/* null command */
#define SEARCHM		0x20		/* enter search mode */
#define RESMCLK		0x40		/* reset missing clock */
#define DISDPLL		0x60		/* disable DPLL */
#define SRCBRG		0x80		/* set source = BR generator */
#define SRCRTXC		0xA0		/* set source = RTxC */
#define FMMODE		0xC0		/* set FM mode */
#define NRZIM		0xE0		/* set NRZI mode */

/*	Write Register 15 - External/Status Interrupt Control */

#define ZCNTIE		0x2		/* zero count interrupt enable */
#define DCDIE		0x8		/* DCD interrupt enable */
#define SYNHTIE		0x10		/* SYNC/HUNT interrupt enable */
#define CTSIE		0x20		/* CTS interrupt enable */
#define TXURNIE		0x40		/* Tx underrun interrupt enable */
#define BRABTIE		0x80		/* break/abort interrupt enable */

/*	Read Register 0 - Transmit/Receive Buffer status & External status */

#define RCA		0x1		/* receive character available */
#define ZCOUNT		0x2		/* zero count */
#define TBE		0x4		/* Tx buffer empty */
#define DCD		0x8		/* data carrier detect */
#define SYNCHT		0x10		/* SYNC/HUNT */
#define CTS		0x20		/* clear to send */
#define TXURN		0x40		/* Tx underrun/EOM */
#define BRABT		0x80		/* break/abort */

/*	Read Register 1 - Special Receive Condition status */

#define ALLSENT		0x1		/* all sent */
#define RESCD2		0x2		/* residue code 2 */
#define RESCD1		0x4		/* residue code 1 */
#define RESCD0		0x8		/* residue code 0 */
#define PERR		0x10		/* parity error */
#define ROVRN		0x20		/* receiver overrun */
#define FRERR		0x40		/* framing error */
#define EOFRM		0x80		/* end of frame */

/* 	Read Register 2B - Interrupt Vector including status */

#define	VECXINT		0x0		/* transmit interrupt */
#define	VECESINT	0x2		/* external status int */
#define	VECRCAINT	0x4		/* received character available int */
#define	VECSRCINT	0x6		/* special receive condition int */
#define	VECCOND		0x6		/* vector condition bit mask */
#define	VECCHA		0x8		/* channel bit mask */
#define VECMSK		(VECCHA|VECCOND)/* masks out all but int status bits */

#define BTXINT		VECXINT		/* ch B transmit int */
#define BESINT		VECESINT	/* ch B external status int */
#define BRCAINT		VECRCAINT	/* ch B received character int */
#define BSRCINT		VECSRCINT	/* ch B special receive condition int*/ 
#define ATXINT		(VECCHA|VECXINT)  /* ch A tx int */
#define AESINT		(VECCHA|VECESINT) /* ch A ext stat int */
#define ARCAINT		(VECCHA|VECRCAINT) /* ch A rec char avail int */
#define ASRCINT		(VECCHA|VECSRCINT) /* ch A special rec cond int */


/*	Read Register 3 - Interrupt Pending register */

#define CHBESIP 	0x1		/* channel B ext/stat int pending */
#define CHBTXIP		0x2		/* channel B Tx int pending */
#define CHBRXIP		0x4		/* channel B Rx int pending */
#define CHAESIP		0x8		/* channel A ext/stat int pending */
#define CHATXIP		0x10		/* channel A Tx int pending */
#define CHARXIP		0x20		/* channel A Rx int pending */

/*	Read Register 10 - Miscellaneous status bits */

#define ONLOOP		0x2		/* on loop (SDLC) */
#define LPSNDING	0x10		/* loop sending */
#define TWOCLKM		0x40		/* two clocks missing */
#define ONECLKM		0x80		/* one clock missing */ 

/*	Read Register 15  - Reflects value stored in WR15 */

/*
 *	Same bit map as WR15
 */

/********** TIME CONSTANT VALUES FOR VARIOUS BAUD RATES ******/

/* Time Constant Values for the MVME147 */

#define VME147_BD50L	0x33
#define VME147_BD50H	0x0C

#define VME147_BD75L	0x21
#define VME147_BD75H	0x08

#define VME147_BD110L	0x8A
#define VME147_BD110H	0x05

#define VME147_BD134L	0x8C
#define VME147_BD134H	0x04

#define VME147_BD150L	0x10
#define VME147_BD150H	0x04

#define VME147_BD200L	0x0B
#define VME147_BD200H	0x03

#define VME147_BD300L	0x07
#define VME147_BD300H	0x02

#define VME147_BD600L	0x02
#define VME147_BD600H	0x01

#define VME147_BD1200L	0x80
#define VME147_BD1200H	0x00

#define VME147_BD1800L	0x55
#define VME147_BD1800H 	0x00

#define VME147_BD2400L	0x3F
#define VME147_BD2400H	0x00

#define VME147_BD3600L	0x29
#define VME147_BD3600H	0x00

#define VME147_BD4800L	0x1F
#define VME147_BD4800H	0x00

#define VME147_BD7200L	0x14
#define VME147_BD7200H 	0x00

#define VME147_BD9600L	0x0E
#define VME147_BD9600H	0x00

#define VME147_BD19200L 0x06
#define VME147_BD19200H 0x00

/* Time Constant values for the MVME134 */

#define VME134_BD50L	0xFF
#define VME134_BD50H	0x02

#define VME134_BD75L	0xFE
#define VME134_BD75H	0x01

#define VME134_BD110L	0x5B
#define VME134_BD110H	0x01

#define VME134_BD134L	0x1D
#define VME134_BD134H	0x01

#define VME134_BD150L	0xFE
#define VME134_BD150H	0x00

#define VME134_BD200L	0xBE
#define VME134_BD200H	0x00

#define VME134_BD300L	0x7E
#define VME134_BD300H	0x00

#define VME134_BD600L	0x3E
#define VME134_BD600H	0x00

#define VME134_BD1200L	0x1E
#define VME134_BD1200H	0x00

#define VME134_BD1800L	0x13
#define VME134_BD1800H 	0x00

#define VME134_BD2400L	0x0E
#define VME134_BD2400H	0x00

#define VME134_BD3600L	0x09
#define VME134_BD3600H	0x00

#define VME134_BD4800L	0x06
#define VME134_BD4800H	0x00

#define VME134_BD7200L	0x03
#define VME134_BD7200H 	0x00

#define VME134_BD9600L	0x02
#define VME134_BD9600H	0x00

#define VME134_BD19200L 0
#define VME134_BD19200H 0

/***************************************************************	
 *
 *	Structure of Z8530 hardware registers
 *
 ***************************************************************/
define(`DEVICE', `z530_io')dnl

struct DEVICE
{
	unsigned char z530_rwr;		/* read/write register */
	unsigned char z530_rtdr;	/* receiver/transmitter data register */
};

/*
 *	Structure of saved registers 
 */

struct z530_sv_reg
{
	unsigned char sv_wr0;
	unsigned char sv_wr1;
	unsigned char sv_wr2;
	unsigned char sv_wr3;
	unsigned char sv_wr4;
	unsigned char sv_wr5;
	unsigned char sv_wr6;
	unsigned char sv_wr7;
	unsigned char sv_wr8;
	unsigned char sv_wr9;
	unsigned char sv_wr10;
	unsigned char sv_wr11;
	unsigned char sv_wr12;
	unsigned char sv_wr13;
	unsigned char sv_wr14;
	unsigned char sv_wr15;
};

/**********  DEFINES FOR RECEIVER AND SCANNER ******************/

#define Z530CSCAN	2	/* clock ticks for character scan */
#define Z530_CHAR	01	/* character arrived flag (z530_work) */
#define Z530_ACT	02	/* character in buffer flag (z530_work */

/******************* EXTERNALS *********************************/
/* Declarations in conf.c */

extern	int	z530_cnt;		/* count of serial ports, conf.c */
extern 	int	z530_ivec[];		/* interrupt vector, from conf.c */
extern 	struct 	tty	z530_tty[];	/* tty structures */
extern 	struct	z530_io	*z530_addr[];	/* from conf.c */
extern 	ttrstrt();
extern  unsigned int  consdev;		/* console device */
extern  unsigned int  dualcons;		/* alternate console device */
extern 	int	Dconcurrent;		/* dual console mode enable */
extern	int	cpuboard;		/* board type */

/******************* DECLARATIONS ********************************/

/* flags which should be in tty structure (t_state) but are not because
   the tty structure must not be enlarged. */
short	z530_state[MAXPORTS];

/* saved write registers */
struct	z530_sv_reg z530_sv_reg[MAXPORTS];

/* declarations for receiver and scanner */
short	z530_scn_flg;
char	z530_work[MAXPORTS];


/* flags for keeping track of DCD and BREAK levels */
short brkflg[MAXPORTS];
short dcdflg[MAXPORTS];

/* speed tables */

unsigned char m147l_speeds[] = {
	0,
	VME147_BD50L,
	VME147_BD75L,
	VME147_BD110L,
	VME147_BD134L,
	VME147_BD150L,
	VME147_BD200L,
	VME147_BD300L,
	VME147_BD600L,
	VME147_BD1200L,
	VME147_BD1800L,
	VME147_BD2400L,
	VME147_BD4800L,
	VME147_BD9600L,
	VME147_BD19200L,
	VME147_BD19200L
};

unsigned char m147h_speeds[] = {
	0,
	VME147_BD50H,
	VME147_BD75H,
	VME147_BD110H,
	VME147_BD134H,
	VME147_BD150H,
	VME147_BD200H,
	VME147_BD300H,
	VME147_BD600H,
	VME147_BD1200H,
	VME147_BD1800H,
	VME147_BD2400H,
	VME147_BD4800H,
	VME147_BD9600H,
	VME147_BD19200H,
	VME147_BD19200H
};

unsigned char m134l_speeds[] = {
	0,
	VME134_BD50L,
	VME134_BD75L,
	VME134_BD110L,
	VME134_BD134L,
	VME134_BD150L,
	VME134_BD200L,
	VME134_BD300L,
	VME134_BD600L,
	VME134_BD1200L,
	VME134_BD1800L,
	VME134_BD2400L,
	VME134_BD4800L,
	VME134_BD9600L,
	VME134_BD19200L,
	VME134_BD19200L
};

unsigned char m134h_speeds[] = {
	0,
	VME134_BD50H,
	VME134_BD75H,
	VME134_BD110H,
	VME134_BD134H,
	VME134_BD150H,
	VME134_BD200H,
	VME134_BD300H,
	VME134_BD600H,
	VME134_BD1200H,
	VME134_BD1800H,
	VME134_BD2400H,
	VME134_BD4800H,
	VME134_BD9600H,
	VME134_BD19200H,
	VME134_BD19200H
};

unsigned char *z530l_speeds, *z530h_speeds;

/******************* DEVICE-SPECIFIC MACROS *********************/
/* 
 * see gtty.m4
 */

define(`GTTYopen',`z530open')dnl
define(`GTTYclose',`z530close')dnl
define(`GTTYread',`z530read')dnl
define(`GTTYwrite',`z530write')dnl
define(`GTTYioctl',`z530ioctl')dnl
define(`GTTYinputc',`z530inputc')dnl
define(`GTTYputc',`z530putc')dnl
define(`GTTYproc',`z530proc')dnl
define(`GTTYmodem',`z530modem')dnl
define(`GTTYparam',`z530param')dnl
define(`GTTYscan',`z530scan')dnl

define(`GTTY_state',`z530_state')dnl
define(`GTTY_cnt',`z530_cnt')dnl
define(`GTTY_delay',`z530_delay')dnl
define(`GTTY_tty',`z530_tty')dnl
define(`GTTY_addr',`z530_addr')dnl
define(`GTTY_scan_flg',`z530_scn_flg')dnl
define(`GTTY_work',`z530_work')dnl
define(`GTTY_debug',`z530_debug')dnl
define(`GTTYMSCAN',`Z530MSCAN')dnl
define(`GTTY_CHAR',`Z530_CHAR')dnl
define(`GTTY_ACT',`Z530_ACT')dnl
define(`GTTYCSCAN',`Z530CSCAN')dnl
define(`DEVICE',`z530_io')dnl
define(`cdelay',`')dnl
define(`GTTY_dcdscn',`')dnl

define(`SPLINT',`spl6()')dnl
define(`SPLSCAN',`spl6()')dnl

include(`gtty.m4')

/* 
 * 	transmitter interrupt enable 
 */

define(`GTTYxie',
		`{
			z530_sv_reg[dev].sv_wr1 |= TXINTEN;
			WRITE_REG(REG1, z530_sv_reg[dev].sv_wr1);
		}')dnl

/*
 *	transmitter interrupt disable 
 */	

define(`GTTYxid',
		`{
			z530_sv_reg[dev].sv_wr1 &= ~TXINTEN;
			WRITE_REG(REG1, z530_sv_reg[dev].sv_wr1);
		}')dnl

/*
 *	receiver interrupt enable 
 */

define(`GTTYrie',
		`{
			z530_sv_reg[dev].sv_wr1 |= RXINTALL;
			WRITE_REG(REG1, z530_sv_reg[dev].sv_wr1);
		}')dnl

/*
 *	receiver interrupt disable
 */

define(`GTTYrid',
		`{
			z530_sv_reg[dev].sv_wr1 &= ~RXINTMSK;
			WRITE_REG(REG1, z530_sv_reg[dev].sv_wr1);
		}')dnl 

/*
 *	output character to transmitter data register 
 */

define(`GTTYoutch',
		`{
			addr->z530_rtdr = $1;
		}')dnl

/*
 *	transmitter enable 
 */

define(`GTTYtxen',
		`{
			z530_sv_reg[dev].sv_wr5 |= TXENAB;
			WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);
		}')dnl

/*
 *	transmitter disable 
 */

define(`GTTYtxdis',
		`{
			z530_sv_reg[dev].sv_wr5 &= ~TXENAB;
			WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);
		}')dnl

/*
 *	transmitter data register available 
 */

define(`GTTYtdra', `(addr->z530_rwr & TBE)')dnl

/*
 *	received data available 
 */

define(`GTTYrda',`(addr->z530_rwr & RCA)')dnl

/*
 *	start sending break
 */

define(`GTTYbrkon',
		`{
			z530_sv_reg[dev].sv_wr5 |= SENDBR;
			WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);
		}')dnl

/*
 *	stop sending break
 */

define(`GTTYbrkoff',
		`{
			z530_sv_reg[dev].sv_wr5 &= ~SENDBR;
			WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);
		}')dnl

/*
 *	initialize dcd and break status 
 */

define(`GTTYinitvars',
		`{
			brkflg[dev] = 0;
			dcdflg[dev] = (addr->z530_rwr & DCD)? G_ON : G_OFF;
		}')dnl
		
/*
 *	reset highest interrupt under service 
 */

define(`GTTYresint',
		`{
			z530_sv_reg[dev].sv_wr0 |= RESHIUS;
			WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0);
			z530_sv_reg[dev].sv_wr0 &= ~RESHIUS;
		}')dnl

/*
 *	reset external status interrupt 
 */

define(`GTTYresesint',
		`{
			z530_sv_reg[dev].sv_wr0 |= RESINT;
			WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0); 
			z530_sv_reg[dev].sv_wr0 &= ~RESINT;
		}')dnl

/*
 *	Output a character and enable transmitter interrupts. On this 
 *	device, the transmitter interrupt has to be enabled before the
 *	character is written to the transmit data register.
 */
	
define(`GTTYoutc_xie',
		`{
			GTTYxie();	/* enable transmitter interrupt */
			GTTYoutch($1);			
		}')dnl


/* 	The macro WRITE_REG takes two arguments.  It writes the second
 * 	argument to the register whose number is specified by the first
 *	argument.  
 */

define(`WRITE_REG',
		`{
			addr->z530_rwr = $1;
			addr->z530_rwr = $2;
		 }')dnl


GTTY_open();

GTTY_close();

GTTY_read();

GTTY_write();

#define	CFLAGMASK	(CBAUD|CSIZE|CSTOPB|PARENB|PARODD|CREAD)

GTTY_ioctl();

GTTY_proc();

GTTY_scan();
/*
 *
 *	param entry
 *
 */

z530param(dev)
unsigned int dev;
{
	register struct tty *tp;
	register struct DEVICE *addr;
	register int flags, indx;
	unsigned char rxcsize, txcsize;
	int cur_lvl;

	LOG(LOG_START, ("z530param: started dev=%d\n", dev));

	GTTY_startscan();

	addr = DEVTOADDR();
	tp = DEVTOTP();
	flags = tp->t_cflag;

	LOG(LOG_INTERN, ("z530param: tp=0x%x, flags=0x%x\n", tp, flags)); 

	if ( dev & 1)
		addr++;


	z530_sv_reg[dev].sv_wr1 &= ~ESINTEN;
	z530_sv_reg[dev].sv_wr3 &= ~RXENAB;
	z530_sv_reg[dev].sv_wr5 &= ~TXENAB;
	z530_sv_reg[dev].sv_wr14 &= ~BRGEN;

	cur_lvl = SPLINT();
	
	WRITE_REG(REG1, z530_sv_reg[dev].sv_wr1);
	WRITE_REG(REG3, z530_sv_reg[dev].sv_wr3);
	WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);
	WRITE_REG(REG14, z530_sv_reg[dev].sv_wr14);

	splx(cur_lvl);

	GTTY_b0();

	GTTY_clocl();


	indx = (flags & CBAUD);

	z530_sv_reg[dev].sv_wr12 = z530l_speeds[indx];
	z530_sv_reg[dev].sv_wr13 = z530h_speeds[indx];
	z530_sv_reg[dev].sv_wr4 &= ~STBMSK;

	if (flags & CSTOPB)
		z530_sv_reg[dev].sv_wr4  |= STBIT2;
	else
		z530_sv_reg[dev].sv_wr4  |= STBIT1;

	if (flags & PARENB) {
		z530_sv_reg[dev].sv_wr1 |= PARSRC;
		z530_sv_reg[dev].sv_wr4 |= ENPAR;

		if (flags & PARODD)
			z530_sv_reg[dev].sv_wr4 &= ~EVENPAR;
		else
			z530_sv_reg[dev].sv_wr4 |= EVENPAR;
	} else {
		z530_sv_reg[dev].sv_wr4 &= ~ENPAR;
	}

	switch (flags & CSIZE) {
	case CS5:	rxcsize = RX5BITS;
			txcsize = TX5BITS;
			break;

	case CS6:	rxcsize = RX6BITS;
			txcsize = TX6BITS;
			break;

	case CS7:	rxcsize = RX7BITS;
			txcsize = TX7BITS;
			break;

	case CS8:	rxcsize = RX8BITS;
			txcsize = TX8BITS;
			break;
	}

	z530_sv_reg[dev].sv_wr3 &= ~RXCHMSK;
	z530_sv_reg[dev].sv_wr3 |= rxcsize;

	z530_sv_reg[dev].sv_wr5 &= ~TXCHMSK;
	z530_sv_reg[dev].sv_wr5 |= txcsize;

	z530_sv_reg[dev].sv_wr14 |= BRGEN;

	if (flags & CREAD) {
		z530_sv_reg[dev].sv_wr3 |= RXENAB;
	}

	z530_sv_reg[dev].sv_wr5 |= TXENAB;
	z530_sv_reg[dev].sv_wr1 |= ESINTEN; 

	cur_lvl = SPLINT();

	WRITE_REG(REG12, z530_sv_reg[dev].sv_wr12);
	WRITE_REG(REG13, z530_sv_reg[dev].sv_wr13);
	WRITE_REG(REG4, z530_sv_reg[dev].sv_wr4);
	WRITE_REG(REG14, z530_sv_reg[dev].sv_wr14);
	WRITE_REG(REG3, z530_sv_reg[dev].sv_wr3);
	WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);

	z530_sv_reg[dev].sv_wr0 |= RESINT;
	WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0);
	z530_sv_reg[dev].sv_wr0 &= ~RESINT;

	WRITE_REG(REG1, z530_sv_reg[dev].sv_wr1);  

	splx(cur_lvl);
	
	LOG(LOG_RETURN, ("z530param: return okay\n"));

} 


/*
 *
 *	init entry
 *
 */

z530init()
{
	register unsigned int dev;
	register struct DEVICE *addr;
	int i, cur_lvl;

	LOG(LOG_WARN, ("Z530_DRIVER: z530_debug=%d, addr=0x%x\n",
	   z530_debug, &z530_debug));

	LOG(LOG_START, ("z530init: started\n"));

	if (cpuboard == VME147) {
		z530l_speeds = m147l_speeds;
		z530h_speeds = m147h_speeds;
	} else {
		z530l_speeds = m134l_speeds;
		z530h_speeds = m134h_speeds;
	}
		
	cur_lvl = SPLINT();

	for (dev = 0; dev < z530_cnt; dev += 2) {

		if (bprobe ( (char *) &(z530_addr[dev]->z530_rwr), -1)) {
			z530_addr[dev] = MISSING;
			z530_addr[dev+1] = MISSING;
			continue;
		}
		addr = DEVTOADDR();

		for (i = 0; i < 2; i++) {

			z530_sv_reg[dev+i].sv_wr4 = STBIT1 | X16CLK;
			if(!Dconcurrent){
				WRITE_REG(REG4, z530_sv_reg[dev+i].sv_wr4); 		
			}

			z530_sv_reg[dev+i].sv_wr12 = z530l_speeds[B9600];
			if(!Dconcurrent){
				WRITE_REG(REG12, z530_sv_reg[dev+i].sv_wr12);
			}

			z530_sv_reg[dev+i].sv_wr13 = z530h_speeds[B9600];
			if(!Dconcurrent){
				WRITE_REG(REG13, z530_sv_reg[dev+i].sv_wr13);
			}

			if (cpuboard == VME147) 
				z530_sv_reg[dev+i].sv_wr14 = BRGEN | BRGSRC;
			else
				z530_sv_reg[dev+i].sv_wr14 = BRGEN;

			WRITE_REG(REG14, z530_sv_reg[dev+i].sv_wr14);

			z530_sv_reg[dev+i].sv_wr2 = z530_ivec[dev/2] >> 2;
			WRITE_REG(REG2, z530_sv_reg[dev+i].sv_wr2);

			z530_sv_reg[dev+i].sv_wr11=BRGOUT|TRXCOI|TCBRGO |RCBRGO;
			WRITE_REG(REG11, z530_sv_reg[dev+i].sv_wr11);

		/*	z530_sv_reg[dev+i].sv_wr10 = 0;
			WRITE_REG(REG10, z530_sv_reg[dev+i].sv_wr10); */

			z530_sv_reg[dev+i].sv_wr3 = RXENAB | RX8BITS;
			WRITE_REG(REG3, z530_sv_reg[dev+i].sv_wr3);	
	
			z530_sv_reg[dev+i].sv_wr5 = TXENAB |RTS |DTR |TX8BITS;
			WRITE_REG(REG5, z530_sv_reg[dev+i].sv_wr5);

			z530_sv_reg[dev+i].sv_wr1 = ESINTEN; 
			WRITE_REG(REG1, z530_sv_reg[dev+i].sv_wr1);
			
			z530_sv_reg[dev+i].sv_wr15 = BRABTIE;
			WRITE_REG(REG15, z530_sv_reg[dev+i].sv_wr15);

			addr++; 
		}

		addr = DEVTOADDR();

		z530_sv_reg[dev].sv_wr9 = MIE;
		WRITE_REG(REG9, z530_sv_reg[dev].sv_wr9);
	}

	splx(cur_lvl);

	LOG(LOG_RETURN, ("z530init: return okay\n"));

}
	
/*
 *
 *	modem entry
 *
 */

z530modem(dev, flag)
unsigned int dev;
int flag;
{
	register struct DEVICE 	*addr;
	register struct tty 	*tp;
	int	cur_lvl, s;

	LOG(LOG_START, ("z530modem: started dev=%d, flag=ox%x\n", dev, flag));

	addr = DEVTOADDR();
	tp = DEVTOTP();
	cur_lvl = SPLINT();

	if ( dev & 1)
		addr++;

	switch (dev)  {

	/*
	 * port 1 does not detect carrier on the mvme134 and the mvme147
	 * both ports detect carrier on the mvme143
	 */
	case PORT1:
		if (cpuboard == VME134 || cpuboard == VME147) {
			s = G_ON;
			break;
		}
	case PORT0:
	case PORT2:
	case PORT3:
		if (flag == G_OFF) {
			z530_sv_reg[dev].sv_wr15 &= ~DCDIE;
			WRITE_REG(REG15, z530_sv_reg[dev].sv_wr15);

			z530_sv_reg[dev].sv_wr5 &= ~(RTS | DTR);
			WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);

			if (!(tp->t_cflag & CLOCAL)) {
				if (tp->t_state & CARR_ON) {
					if (tp->t_state & ISOPEN) {
						signal(tp->t_pgrp, SIGHUP);
						ttyflush(tp, (FREAD | FWRITE));
					}
					tp->t_state &= ~CARR_ON;
				}
			}
			else if (!(tp->t_state & CARR_ON)) {
				wakeup(&tp->t_canq);
				tp->t_state |= CARR_ON;
			}
		} 
		else if (flag == G_ON) {
			z530_sv_reg[dev].sv_wr5 |= RTS | DTR;
			WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);

			z530_sv_reg[dev].sv_wr15 |= DCDIE;
			WRITE_REG(REG15, z530_sv_reg[dev].sv_wr15);
		}

		s = addr->z530_rwr & DCD;
	}
	
	splx(cur_lvl);

	LOG(LOG_INTERN, ("z530modem: s=%d\n", s));

	LOG(LOG_RETURN, ("z530modem: return okay\n"));

	return(s);

}


/*
 *
 *	intr entry
 *
 */

z530intr(dev)
unsigned int	dev;
{
	register struct	DEVICE	*addr;
	register struct tty	*tp;
	register unsigned char	c;
	register int	flg, count;
	char	lbuf[3];
	int	lcnt;
	int	cur_lvl;	
	unsigned char ipstat, errstat, extstat;
	unsigned int scc, port;
	char discard;
	unsigned char dcstatus;

	LOG(LOG_START, ("z530intr: started dev=%d\n", dev));

	addr = DEVTOADDR();
	if (addr == MISSING)
		return;


	addr->z530_rwr = REG2;			/* read int vector from ch. B */
	ipstat = addr->z530_rwr & VECMSK;	/* to get the modified vector */

	if (ipstat & VECCHA) {
		dev = dev + 1;
		++addr;
	}

	LOG(LOG_INTERN, ("z530intr: dev = %d, ipstat = 0x%x, addr = 0x%x\n", 
	dev, ipstat, addr));

	tp = DEVTOTP();
	 
	switch (ipstat) {
	
	case ATXINT:
	case BTXINT:

		LOG(LOG_INTERN, ("z530intr: TXINT, dev = %d\n", dev));

		GTTY_xint();


		if (!(tp->t_state & BUSY)) { 
			z530_sv_reg[dev].sv_wr0 |= RESTXIP;
			WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0);
			z530_sv_reg[dev].sv_wr0 &= ~RESTXIP;
		}


		z530_sv_reg[dev].sv_wr0 |= RESHIUS;
		WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0);
		z530_sv_reg[dev].sv_wr0 &= ~RESHIUS;

		LOG(LOG_INTERN, ("z530intr: return okay\n"));
		return;

	case ARCAINT:
	case BRCAINT:

		c = addr->z530_rtdr;

		if(Dconcurrent)
			if(z530_state[dev] & DCON){
				dcstatus = 0;
				consinputc(c, dcstatus);
			}

		LOG(LOG_INTERN, ("z530intr: RCAINT, char=0x%x\n", c));
	
		lcnt = 1;
		flg = tp->t_iflag;
		GTTY_rint();
		break;

	case ASRCINT:
	case BSRCINT:

		addr->z530_rwr = REG1;
		errstat = addr->z530_rwr;
		c = addr->z530_rtdr;

		if(Dconcurrent)
			if(z530_state[dev] & DCON){
				dcstatus = 0;
				if(errstat & PERR)
					dcstatus |= DC_PARERR;
				if(errstat & ROVRN)
					dcstatus |= DC_ROVRN;
				if(errstat & FRERR)
					dcstatus |= DC_FRERR;
				consinputc(c, dcstatus);
			}

		LOG(LOG_INTERN, ("z530intr: SRCINT, char=0x%x\n", c));

		z530_sv_reg[dev].sv_wr0 |= ERRRES;
		WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0);
		z530_sv_reg[dev].sv_wr0 &= ~ERRRES;

		lcnt = 1;
		flg = tp->t_iflag;
		
		if (!(flg & IGNPAR)) {
			if (!(flg & INPCK))
				errstat &= ~PERR;
			if (errstat & (PERR | ROVRN | FRERR))
				GTTY_eint();
		}
		break;

	case AESINT:
	case BESINT:

		LOG(LOG_INTERN, ("z530intr: ESINT\n"));

		extstat = addr->z530_rwr;
		flg = tp->t_iflag;
	

		/*
		 * port 1 does not detect carrier on the mvme134 and mvme147
		 * both ports detect carrier on the mvme143
	 	*/
		if (dev != 1 || cpuboard == VME143) {
			if (dcdflg[dev]) {
				if (!(extstat & DCD)) {
					dcdflg[dev] = 0;
					LOG(LOG_INTERN, 
					("z530intr: ESINT dcd off\n"));
					GTTY_dcdoffint();
					if(Dconcurrent && (z530_state[dev] & DCON))
						consinputc(0, DC_DCDOFF);
				}
			}
			else if (extstat & DCD) {
				dcdflg[dev]++;
				LOG(LOG_INTERN, 
				("z530intr: ESINT dcd on\n"));

				GTTY_dcdonint();
				if (Dconcurrent && (z530_state[dev] & DCON))
					consinputc(0, DC_DCDON);
			}
		}

		if (brkflg[dev] ) {
			if (!(extstat & BRABT)) {
				brkflg[dev] = 0;
				discard = addr->z530_rtdr;
				LOG(LOG_INTERN,
				("z530intr: ESINT end of break\n"));
			}
		}
		else if (extstat & BRABT) {
			brkflg[dev]++;
			flg = tp->t_iflag;
			LOG(LOG_INTERN, 
			("z530intr: ESINT start of break\n"));

			if(Dconcurrent && (z530_state[dev] & DCON)){
				dcstatus = DC_BREAK;
				c = 0;
				consinputc(c, dcstatus);
			}
			GTTY_bint();
		}

		z530_sv_reg[dev].sv_wr0 |= RESINT;
		WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0); 
		z530_sv_reg[dev].sv_wr0 &= ~RESINT;
		

		z530_sv_reg[dev].sv_wr0 |= RESHIUS;
		WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0);
		z530_sv_reg[dev].sv_wr0 &= ~RESHIUS;

		LOG(LOG_INTERN, ("z530intr: return okay\n"));

		return;
	}	/* end of switch */

	/* stash character in r_buf */
	
	GTTY_rbuf();

	z530_sv_reg[dev].sv_wr0 |= RESHIUS;
	WRITE_REG(REG0, z530_sv_reg[dev].sv_wr0);
	z530_sv_reg[dev].sv_wr0 &= ~RESHIUS;

	LOG(LOG_INTERN, ("z530intr: return okay\n"));

	
}
/*
 *
 *  inputc
 *
 */

z530inputc(dev, c, rxstat)
register unsigned int	dev;
register unsigned char	c;
register unsigned char	rxstat;
{
	register struct	DEVICE	*addr;
	register struct tty	*tp;
	register int	flg,count;
	char	lbuf[3];
	int	lcnt;
	int	cur_lvl;

	LOG(LOG_INTERN, ("z530inputc: char=0x%x\n", c));
	
	tp = DEVTOTP();
	flg= tp->t_iflag;
	
	switch (rxstat) {

	case DC_BREAK:
		if ( !(flg & IGNBRK) && (flg & BRKINT) ) {
			(*linesw[tp->t_line].l_input) (tp, L_BREAK);
		}
		return;

	case DC_PARERR:
	case DC_ROVRN:
	case DC_FRERR:
		
		lcnt = 1;
		flg = tp->t_iflag;
		
		if (!(flg & IGNPAR)) {
			if (!(flg & INPCK))
				rxstat &= ~DC_PARERR;
			if (rxstat & (DC_PARERR | DC_ROVRN | DC_FRERR)) {
				/*
		 		 *	If PARMRK is set, a character with a framing
		 		 *	or parity error is read as the three-character
		 		 *	sequence 0xFF, 0, X, where X is the data of the
		 		 *	character received in error. 
		 		 */

				if (flg & PARMRK) {					
					lbuf[2] = 0xFF;						
					lbuf[1] = 0;					
					lcnt = 3;					
					sysinfo.rawch += 2;				
				} 
				/*
		 		 *	If PARMRK (mark parity errors) is not set,
		 		 *	a framing or parity error which is not 
		 		 *	ignored is read as the character NUL (0).
		 		 */

				else c = 0;						
			}
		}
		break;
		
	case DC_DCDON:
		GTTY_dcdonint();
		return;

	case DC_DCDOFF:
		if (!(tp->t_cflag & CLOCAL)) { 
			if (tp->t_state & CARR_ON) {				
				if (tp->t_state & ISOPEN) {			
			    		signal(tp->t_pgrp, SIGHUP);
			    		ttyflush(tp, (FREAD | FWRITE));		
				}						
				tp->t_state &= ~CARR_ON;			
			}
		}							
		return;

	default:
		lcnt = 1;
		flg = tp->t_iflag;
		GTTY_rint();
		break;
	}

	/* stash character in r_buf */
	GTTY_rbuf();

	LOG(LOG_INTERN, ("z530inputc: return okay\n"));
}

	
/*
 *
 *	putc entry
 *
 */


z530putc(dev, c)
register unsigned int dev;
register char c;
{
	register struct DEVICE *addr;
	register struct tty *tp;
	int cur_lvl;

	dev = minor(dev);
	addr = DEVTOADDR();
	INCADDR();
	cur_lvl = SPLINT;
	
	WRITE_REG(REG9, ( z530_sv_reg[dev].sv_wr9 & ~MIE ));
	WRITE_REG(REG5, ( z530_sv_reg[dev].sv_wr5 | TXENAB ));

	for (;;)
	{
		while (!(GTTYtdra()));
		addr->z530_rtdr = c;
		if (c != LF)
			break;
		c = CR;
	}

	addr->z530_rwr = REG1;
	while (!(addr->z530_rwr & ALLSENT))
		addr->z530_rwr = REG1;

	WRITE_REG(REG5, z530_sv_reg[dev].sv_wr5);  /* restore previous state */
	WRITE_REG(REG9, (z530_sv_reg[dev].sv_wr9 | MIE));

	tp = DEVTOTP();
	if (tp->t_state & TTXOFF) {
		GTTYoutch(CSTOP);
		tp->t_state &= ~TTXOFF;
	} else if (tp->t_state & TTXON) {
		GTTYoutch(CSTART);
		tp->t_state &= ~TTXON;
	} else  if (tp->t_state & BUSY) {
			tp->t_state &= ~BUSY;
			z530proc(tp, T_OUTPUT);
 	}

	splx(cur_lvl);
}

	
