#
#	Copyright (c) 1985 AT&T
#	Copyright (c) 1988 Motorola Inc
#	All Rights Reserved
#	@(#)trap.s	35.5

#####################################################################
#		File:  trap.s
#
#  This package provides exception handling for Unix.  Basically,
#  exceptions are divided into one of several basic types:
#
#  1)  Interrupts (Notably the clock)
#  2)  System calls.
#  3)  Traps (divide by 0, illegal instructions, etc.)
#  4)  Bus Errors (handled in paging.s)
#  5)  Profiling clock interrupts
#
#####################################################################

set	SAVEREGS,0xFFFE		# MOVM pattern for saving all except sp.
set	GETREGS,0x7FFF		# MOVM pattern for getting all except sp.

#####################################################################
# Clock interrupt
#
#   Control comes here whenever a clock interrupt occurs.
#   We need to call clock as follows:
#
#	clock(pc, ps);
#
#   The ps must be the ps of the machine before the interrupt.
#	This is not true on a throw-away stack frame.
#####################################################################
	text
	global	Xclock
Xclock:
	movm.l	&0xc0c0,-(%sp)		# Save the scratch registers.
	mov.w	0x10(%sp),%d1		# Save current status register.
	mov.w	0x16(%sp),%d0		# Type/offset from interrupt stack
	lsr.w	&12,%d0			# shift off vector offset, leave type
	beq.b	normal			# If normal int frame, use above sr.
					# Throw away stack frame!
	or.w	&0x1000,%sr		# switch back to master stack
	mov.w	(%sp),%d1		# get previous status register
	and.w	&0xefff,%sr		# switch back to interrupt stack
normal:
	clr.l	-(%sp)			# Make room for sr on the stack
	mov.w	%d1,2(%sp)		# push sr as arg to clock
	mov.l	16+4+2(%sp),-(%sp)	# PC on the stack.

	jsr	clock			# Call the clock routine.

	add.l	&8,%sp			# Pull SR and PC off.
	tst.l	%d0			# Test return value from clock
	beq.b	intret			# Normal interrupt return

#####################################################################
#
# At this point, clock returned a non-zero value and we need to run the user
# profiling (addupc).  Clock could not run user profiling because it was
# executing on the interrupt stack and addupc can get a page fault and sleep.
# We were previously in user mode before the interrupt, therefore we are the 
# only interrupt on the stack.  To run the user profiling, we must clean up the
# interrupt stack, switch back to the master stack and lower the interrupt
# level before executing addupc(pc, &u.u_prof, 1).
#
#####################################################################

	movm.l	(%sp)+,&0x303		# Restore the scratch registers
	add.l	&0x8,%sp		# Bump past the sr/pc/vector
	or.w	&0x1000,%sr		# Switch back to the master stack
	and.w	&0x3000,%sr		# lower the interrupt level to zero

	movm.l	&0xc0c0,-(%sp)		# Save the scratch registers again.
	mov.l	&1,-(%sp)		# push tick (1) argument
	mov.l	&u+u_prof,-(%sp)	# push u.u_prof address
	mov.l	26(%sp),-(%sp)		# Put the pc on the stack.

	jsr	addupc			# Do the dirty deed

	add.l	&12,%sp			# Pop off args
	movm.l	(%sp)+,&0x303		# Restore the scratch registers
	bra.b	resched			# See if rescheduling is needed

#####################################################################
#
# Common interrupt return for all interrupt routines
#
# Interrupts use the interrupt stack.  It is important that this
# stack gets cleaned up if we are going to switch to another process.
# Process switching can only occur on the master stack.  If we have 
# nested interrupts (normal stack frame), then don't process switch.
# Just return from interrupt to the previous interrupt.
#
#####################################################################

	global	intret
intret:
	jsr     dcachclr
	mov.w	0x16(%sp),%d0	# Type/offset from interrupt stack
	lsr.w	&12,%d0		# shift off vector offset, leave type
	movm.l	(%sp)+,&0x303	# Get back the saved registers
	beq.b	nosched		# If normal interrupt frame, return from int.
				# Another interrupt is underneath this one.

# At this point, we know we were previously using the master stack 
# and the real stack frame is on that stack.  We are going to clean
# up the interrupt stack entirely and switch back to the master stack.

	add.l	&0x8,%sp	# Bump past the sr/pc/vector
	or.w	&0x1000,%sr	# Switch back to the other stack

	global	resched
resched:			# Determine if rescheduling is neccessary.

# Here we try to simulate a software interrupt.  After all pending
# interrupts are serviced and we are going to return to level 0,
# we drop the mask, to allow any low priority interrupts to be 
# serviced, switch to the interrupt stack, and call the timer interrupt.
	tst.l	timeflag	# check if we should run timein()
	beq.b	nottime		# No, nothing to run
	bftst	(%sp){&5:&3}	# Yes, but are we going to level 0?
	bne.b	nottime		#

	clr.l	timeflag	#
	mov.w	&0x2000,%sr	# use interrupt stack and level 0
	movm.l	&0xc0c0,-(%sp)	# Save the scratch registers again.
	jsr	timein		#
	movm.l	(%sp)+,&0x303	# Restore the scratch registers
	or.w	&0x1000,%sr	# Switch back to the master stack
nottime:

	btst	&5,(%sp)	# If interrupted in system mode,
	bne.b	nosched		# then don't reschedule.
	tst.b	runrun		# If runrun is clear, then don't
	beq.b	nosched		# reschedule.
#
# Rescheduling is needed if control comes here.
# We need to call trap with 256 for vecnum to reschedule.
#
	mov.l	2(%sp),-(%sp)	# Put the pc on the stack.
	clr.l	-(%sp)		# Make a space for status.
	mov.w	&256,(%sp)	# Put vector number in status.
	mov.w	8(%sp),2(%sp)	# Put status register in status.
	movm.l	&0xFFFE,-(%sp)	# Pass all user registers.
	mov.l	%usp,%a0	# Save usp on stack
	mov.l	%a0,-(%sp)

	jsr	trap		# Call trap to reschedule it.

	mov.l	(%sp)+,%a0	# restore the usp
	mov.l	%a0,%usp
	movm.l	(%sp)+,&0x7fff	# Recover registers and return.
	mov.w	2(%sp),8(%sp)	# Copy back the status register
	mov.l	4(%sp),10(%sp)	# Copy back the pc
	add.l	&8,%sp		# Bump past the sr/pc
nosched:
	jsr     dcachclr	# Clear chip data cache.
	rte			# return to interrupted process


#####################################################################
#
#			Xtrap - All trap calls
#
#  trap(usp,d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,status,pc);
#	All arguments are long, aligned on 4 byte boundary.
#
#  Control comes here on any exception (addr errors, illegal op., 0 divide etc)
#  Prepare parameter list for trap, it will do the rest.
#  Note that extra lengths are gone to to always maintain the stack 
#  aligned on 4 byte boundaries, so be cautious about making changes. 
#
#####################################################################

nullvect:	global	nullvect
alltraps:	global	alltraps
Xtrap:		global	Xtrap

	mov.l	2(%sp),-(%sp)		# Put the pc on the stack.

	clr.l	-(%sp)			# Make a space for status.
	mov.w	4+4+6(%sp),(%sp)	# Put vector number in status.
	and.w	&0xfff,(%sp)		# Clear format bits if any.
	mov.w	4+4(%sp),2(%sp)		# Put status register in status.

	movm.l	&SAVEREGS,-(%sp)	# Pass all user registers.
	mov.l	%usp,%a0		# Include the usp.
	mov.l	%a0,-(%sp)

	jsr	trap			# Finished w/ parameters, make the call.

	lea.l	76(%sp),%a2		#  Address of type/offset long + 1/2 pc
	mov.l	(%a2),%d0		#  Type/offset of frame to be saved
	lsr.w	&12,%d0			#  Shift out vector offset, leave type

	cmp.w   %d0,&1                  #  Check for throw away stack frame
	beq     L%twf			#  Yes, skip most of this stuff

	mov.b	(stkfmtsz,%d0.w),%d0	#  Using type; lookup frame size (bytes)

#  Remove the stack frame and replace by standard 4 word frame.
#  If the flag is set, save the old frame in the U-block
	mov.l   saveflag,%a0            # Address of flag
	tst.b   (%a0)                   # Test the flag
	beq.b   L%nosav                 # Do not save stack frame
	clr.b   (%a0)                   # Clear the flag
	mov.l   saveaddr,%a0            # Get address of save area
	mov.l   %a2,%a1
	mov.w   %d0,%d1                 # Get size of frame
	lsr.w   &2,%d1                  # Convert bytes to longs
L%loop: mov.l   (%a1)+,(%a0)+           # Move to save area
	dbra.w  %d1,L%loop

L%nosav:
	mov.l	%a2,%a1			#  stack frame address
	add.w	%d0,%a1			#  End of stack frame
	clr.l	(%a1)			#  New Type/offset is zero
	mov.w	&18,%d1			#  15 regs, usp, 4 wd frame, pc/status
L%dlp2:	mov.l	-(%a2),-(%a1)		#  Move registers, usp, status, pc and
	dbra.w	%d1,L%dlp2		#  short frame to end of kernel stack.
	add.w	%d0,%sp			#  Adjust stack by size of stack frame
L%twf:
	mov.l	(%sp)+,%a0		# Restore the users
	mov.l	%a0,%usp		# stack pointer;
	movm.l	(%sp)+,&GETREGS		# and all registers.
	mov.w	2(%sp),4+4(%sp)		# Copy back the status register
	mov.l	4(%sp),4+4+2(%sp)	# Copy back the pc
	add.l	&8,%sp			# Zap status, and pc.
	jsr     dcachclr		# Clear chip data cache.
	rte

#####################################################################
#
#		Xsyscall - All system calls
#
# syscall(usp,d0,d1,d2,d3,d4,d5,d6,d7,a0,a1,a2,a3,a4,a5,a6,status,pc);
#	All arguments are long, aligned on 4 byte boundary.
#
#  Control comes here to process a system call.
#  A system call is generated by a trap #0 instruction.
#  Basically, all we have to do is call syscall() with its
#  parameter list and let it handle the details.
#  Note that extra lengths are gone to always maintain the stack 
#  aligned on 4 byte boundaries, so be cautious about making changes. 
#
#####################################################################
	global	Xsyscall
	global	a45j
	global	cputype
	global	fixprr
Xsyscall:
	mov.l	2(%sp),-(%sp)		#  Put the pc on the stack.

	clr.l	-(%sp)			#  Make a space for status.
	mov.w	4+4+6(%sp),(%sp)	#  Put vector number in status.
	and.w	&0xfff,(%sp)		#  Clear format bits if any.
	mov.w	4+4(%sp),2(%sp)		#  Put status register in status.

	movm.l	&SAVEREGS,-(%sp)	#  Save the registers.
	mov.l	%usp,%a0		#  Including the usp.
	mov.l	%a0,-(%sp)

	jsr	syscall			#  The parameters are built, now do it.

	mov.l	saveflag,%a0		#  Restore stack frame flag address
	tst.b	(%a0)			#  get the flag itself
	beq.b	L%skip2			#  If zero, nothing to do.
					#  else restore the stack frame
	mov.b   (%a0),%d2               #  save flag for prr check later
	clr.b	(%a0)			#  Reset the save-flag
	mov.l	saveaddr,%a0		#  Get address of stack frame save area
	mov.l	(%a0),%d1		#  Get stack frame type/vector offset
	lsr.w	&12,%d1			#  Shift out vector offset, leave type
	mov.b	(stkfmtsz,%d1.w),%d1	#  Using type, lookup frame size (bytes)
	mov.l	%sp,%a2			#  Load current kernel stack pointer
	sub.w	%d1,%sp			#  Subtract frame size to make room
	mov.l	%sp,%a1			#  Save result, this is new top of stack

	mov.w	&18,%d0			#  (15 regs+usp)*4 + (4 wd frame)*2 + 8
L%dlp3:	mov.l	(%a2)+,(%a1)+		#  Move registers to make room for frame
	dbra.w	%d0,L%dlp3		#  New frame is place at end of stack

	lsr.w	&2,%d1			#  Convert bytes to longs
L%dlp4:	mov.l	(%a0),(%a1)+		#  Move saved frame onto kernel stack
	clr.l	(%a0)+			#  Clear the save buffer (invalidated)
	dbra.w	%d1,L%dlp4		#  This should abut end of U-block

	cmp.b   %d2,&2                  #  Retrieve flag saved in d2.  Do not
	beq.b   L%skip2                 #  call fixprr if this is FPE.

	btst	&2,9+72(%sp)		# Check PRR bit. 
	bne.b	L%skip2			# If 1, no fixup is needed.
					# This goes away on good 020 chip.
	tst.w	a45j			# if a45j is zero then don't
	beq.b	L%skip2			# need to fix this on an 020 chip.

	mov.w	&682,%d0		# make sure that we're only
	cmp.w	%d0,cputype		# fixing this on an 020 chip.
	bne.b	L%skip2

	lea.l	72(%sp),%a2		#  pointer to stack frame
	bsr.l	fixprr			#  Fix PRR per errata sheet in buserr.

L%skip2: mov.l	(%sp)+,%a0		#  Restore the user's
	mov.l	%a0,%usp		#  stack pointer.
	movm.l	(%sp)+,&GETREGS		#  Restore all user registers.

	mov.w	2(%sp),4+4(%sp)		#  Copy back the status register
	mov.l	4(%sp),4+4+2(%sp)	#  Copy back the pc
	add.l	&8,%sp			#  Undo status/pc save.
	jsr     dcachclr		# Clear chip data cache.
	rte				#  We're done, bye.

#####################################################################

#####################################################################
# Profiling clock interrupt
#
#   Control comes here whenever a prfclock interrupt occurs.
#   We need to call prfintr as follows:
#
#	prfintr(pc, ps);
#
#   The ps must be the ps of the machine before the interrupt.
#	This is not true on a throw-away stack frame.
#   This is the same crap that Xclock must do, except for addupc stuff
#####################################################################
	text
	global	Xprfclk
Xprfclk:
	movm.l	&0xc0c0,-(%sp)		# Save the scratch registers.
	mov.w	0x10(%sp),%d1		# Save current status register.
	mov.w	0x16(%sp),%d0		# Type/offset from interrupt stack
	lsr.w	&12,%d0			# shift off vector offset, leave type
	beq.b	doit			# If normal int frame, use above sr.
					# Throw away stack frame!
	or.w	&0x1000,%sr		# switch back to master stack
	mov.w	(%sp),%d1		# get previous status register
	and.w	&0xefff,%sr		# switch back to interrupt stack
doit:
	clr.l	-(%sp)			# Make room for sr on the stack
	mov.w	%d1,2(%sp)		# push sr as arg to clock
	mov.l	16+4+2(%sp),-(%sp)	# PC on the stack.

	jsr	prfintr			# Call the profiling routine.

	add.l	&8,%sp			# Pull SR and PC off.
	bra.w	intret			# Normal interrupt return
	nop				# Force non-zero branch to intret
	
#####################################################################

	global	sysint
	global	update
	global  printf
	data	2
sysmes17:
	byte	0x07,'\n,'F,'V,'M,'E,'2,'5,'2,':,0x20
	byte	'A,'C,'F,'A,'I,'L,0x20,'a
	byte	's,'s,'e,'r,'t,'e,'d,0x07
	byte	'\n,0x00
sysmes22:
	byte	0x07,'\n,'F,'V,'M,'E,'2,'5,'2,':,0x20
	byte	'S,'Y,'S,'F,'A,'I,'L,0x20
	byte	'a,'s,'s,'e,'r,'t,'e,'d
	byte	0x07,'\n,0x00
sysmes27:
	byte	0x07,'\n,'F,'V,'M,'E,'2,'5,'2,':,0x20
	byte	'A,'s,'s,'e,'r,'t,'i,'n
	byte	'g,0x20,'o,'n,'e,0x20,'o,'f
	byte	0x20,'t,'h,'e,0x20,'V,'M,'E
	byte	'b,'u,'s,0x20,'I,'N,'T,'R
	byte	'E,'Q,0x20,'l,'i,'n,'e,'s
	byte	0x07,'\n,0x00
sysmes32:
	byte	0x07,'\n,'F,'V,'M,'E,'2,'5,'2,':,0x20
	byte	'A,'r,'b,'i,'t,'r,'a,'t
	byte	'i,'o,'n,0x20,'t,'i,'m,'e
	byte	'o,'u,'t,'!,0x07,'\n,0x00
sysmes34:
	byte	0x07,'\n,'F,'V,'M,'E,'2,'5,'2,':,0x20
	byte	'U,'n,'k,'n,'o,'w,'n,0x20
	byte	'l,'e,'v,'e,'l,0x20,'7,0x20
	byte	'i,'n,'t,'e,'r,'r,'u,'p
	byte	't,0x20,'%,'x,'\n,0x07,0x00
sysmes40:
	byte	0x07,'\n,'F,'V,'M,'E,'2,'5,'2,':,0x20
	byte	'A,'C,'F,'A,'I,'L,0x20,'a,'n,'d,0x20
	byte	'S,'Y,'S,'F,'A,'I,'L,0x20
	byte	'a,'s,'s,'e,'r,'t,'e,'d,'\n,0x07,0x00
	text

	# Level 7 interrupt handler for FVME252
	# Author M. Woods Date 31/8/88

sysint:
	#mov.b	-1048575,%d0	# Read Level 7 interrupt control register
	movm.l	&0x0004, -(%sp)
	mov.b	0xfff00001,%d0	# Read Level 7 interrupt control register
	extb.l	%d0
	and.l	&15, %d0		# Mask out bits 4 - 7
	btst.b  &0,%d0
	bne	sysint15
	btst.b	&1,%d0
	bne	sysint21
	btst.b  &2,%d0
	bne	sysint26
	btst.b	&3,%d0
	bne	sysint27
	cmp.l	%d0, &3
	beq	sysint28
	bne.w	sysint33
sysint27:
	mov.l	&sysmes32,(%sp)	# Arbitration timeout occurred
	jsr	printf
sysintxx:
	bra.b	sysintxx	# Loop forever
sysint28:
	mov.l	&sysmes40, (%sp) # SYSFAIL AND ACFAIL
	jsr	printf
	bra.b	sysintxx
sysint15:
	mov.l	&sysmes17,(%sp)	# ACFAIL interrupt
	jsr	printf
sysint20:
	bra.b	sysint20	# Loop forever
sysint21:
	mov.l	&0x1000000, %d2
sysfail_loop:
	mov.b	0xfff00001,%d0	# Read Level 7 interrupt control register
	extb.l	%d0
	and.l	&15, %d0		# Mask out bits 4 - 7
	btst.b	&1, 0xfff00001
	beq	level7_ret
	sub.l	&1, %d2
	bne	sysfail_loop
	mov.l	&sysmes22,(%sp)	# SYSFAIL interrupt
	jsr	printf
sysint25:
	bra.b	sysint25	# Loop forever
sysint26:
	mov.l	&sysmes27,(%sp)	# FVME252 asserting a INTREQ line.
	jsr	printf
sysint30:
	bra.b	sysint30	# Loop forever
sysint33:
	mov.l	%d0,(%sp)
	mov.l	&sysmes34,-(%sp)
	jsr	printf
	add.l	&4,%sp
sysint37:
	bra.b	sysint37	# Loop forever
level7_ret:
	movm.l	(%sp)+ , &0x0004
	rte
