#ident "@(#)ma_init.c	35.2"
/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	ma_init.c
**
**  DESCRIPTION:	This module will hold all the MAC initialization
**			routines.
**
**  CONTENTS:		ma_init()
**			ma_int_mem()
**			ma_int_rng()
**			ma_int_lnc()
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:
**
**********************************************************************/

/*
**	Include Files
*/

#include <sys/osi/common/types.h>	/* Abstract Type Definitions */
#include <sys/osi/common/system.h>	/* System Function Definitions */
#include <sys/osi/common/bdb.h>		/* BDB Type Definitions */
#include <sys/osi/common/epb.h>		/* EPB Type Definitions */
#include <sys/osi/common/selsizes.h>	/* MAC Address Size */
#include <sys/osi/common/ma_epb.h>	/* MAC EPB Type Definitions */
#include <sys/osi/common/macros.h>	/* Macro Definitions */
#include <sys/osi/common/com_env.h>	/* CE Definitions */
#include <sys/osi/board/mvme147.h>	/* MVME147 NVRAM Definitions */

#include <sys/nvram.h>			/* R3V4 & Later NVRAM Access */

#include "ma_lance.h"			/* LANCE Controller Definitions */
#include "ma_def.h"			/* MAC Node Definitions */
#include "ma_globals.h"			/* Global MAC Node Definitions */

/*
**	External Functions
*/

extern	VOID	ma_isr_asm();	/* LANCE Interrupt Service Routine */
extern	VOID	ma_rx_gbuf();	/* Get RX Buffer */
extern	VOID	ma_set_rap();	/* Set Address Register */
extern	VOID	ma_set_rdp();	/* Set Data Register */
extern	USHORT	ma_get_rdp();	/* Get Data Register */
extern	VOID	o_clean();	/* Clear Bytes */
extern	EPB	*o_getepb();	/* Get An EPB */
extern	BDB	*o_getbdb();	/* Get BDB Block */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ma_init
**
**  DESCRIPTION:	This function performs MAC node initialization,
**			incorporating default parameters and values to
**			initiate LANCE processing.
**
**  PARAMETERS:		
**
**	Input:
**
**	Output:
**
**	Returns:
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ma_init()

{
	VOID	ma_int_mem(),	/* Obtain LANCE Memory */
		ma_int_lnc();	/* LANCE Self-Initialization */

	register USHORT	i;		/* Counter */

	ULONG	retcode;		/* Storage For Return Code */

/*
**  First, the program explicitly stops the LANCE's 
**  operations by setting the STOP bit in LANCE CSR0.
**  This insures integrity in subsequent LANCE
**  initialization.
*/

	retcode = RET_OK;

	ma_set_rap(RAP_CSR0);

	ma_set_rdp(CSR0_STOP);

/*
**  Next, memory is obtained for the LANCE data structures. This
**  includes the LANCE Init. Block, Transmit Desc. Ring, Receive
**  Desc. Ring and the registration table.
**
**  Also the LANCE descriptor rings are initialized. This
**  procedure includes clearing all ring entries and loading the
**  Init Block field related to the ring.
*/

	ma_int_mem(&retcode);
			
/*
**  If no errors have occurred, then we will initialize the
**  LANCE according to configurable and default parameters.
*/

	/*
    	**  First, the default mode of operations is loaded; this mode
    	**  has the receiver and transmitter turned on, and all other
    	**  features turned off.
    	*/

	if (retcode == RET_OK)
	{
		ma_lnc_iblkptr->mode = MODE;

		/*
    		**  Next, the node's MAC address is moved from NVRAM to the 
    		**  global MAC variable ma_ts_addr. When ma_int_lnc() is called,
		**  the node's MAC address is moved from ma_ts_addr to the 
		**  Initialization Block.
    		*/

		if (nvram(READ, NET, ma_ts_addr, MVME147_TS_ADDR_SIZE, MVME147_TS_ADDR_INDEX) < 0)
		{
			retcode = MA_ERR_INT;

			LOG("MAC: This_Station_Address checksum error\n");

			return;
		}

		LOG("MAC: Station address = 0x");

		for (i = 0; i < MACSZ; ++i)
			LOG("%x", ma_ts_addr[i]);

		LOG("\n");

		/*
    		**  Next, the node's Group/Logical address mask is moved from
		**  NVRAM to the global MAC variable ma_ga_mask. When
		**  ma_int_lnc() is called, the address mask is moved from
		**  ma_ga_mask to the Initialization Block.
    		*/

		if (nvram(READ, NET, ma_ga_mask, MVME147_GA_MASK_SIZE, MVME147_GA_MASK_INDEX) < OK)
		{
			LOG("MAC: Group_Address_Mask checksum error, ");
			LOG("default is 0x0000000000000000\n");

			o_clean(ma_ga_mask, MVME147_GA_MASK_SIZE);
		}

		/*
		**  Now the ma_int_lnc() function is called to fill the
		**  LANCE Init Block and to request the LANCE to initialize
		**  initialize itself and begin normal, interrupt-driven
		**  operation.
		*/

		LOG("MAC: LANCE interface error, when logging stops at this place\n");

		ma_int_lnc();

	/*
	**  Upon return, the LANCE CSR0 is read to insure that
	**  the LANCE is running with transmitter and receiver
	**  on and interrupts enabled. The return code is set
	**  to indicate an error if this is not the case.
	*/

		if (ma_get_rdp() != (USHORT)CSR0_GOOD)
		{
			retcode = MA_ERR_INT;

			LOG("MAC: LANCE initialization error\n");
		}
	}

/*
**  Next, if any initialization errors have occurred, we set ma_state
**  to MAC_LOCAL.
*/

	if (retcode != RET_OK)
	{
		LOG("MAC: MAC initialization error, ");
		LOG("node restricted to local activities\n");

		ma_state = MAC_LOCAL;
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ma_int_mem
**
**  DESCRIPTION:	This function will obtain memory for the
**			LANCE data structures. The necessary
**			structures are the LANCE Initialization
**			block, the LANCE Receive Descriptor Ring,
**			the LANCE Transmit Descriptor Ring, and the
**			registration table.
**
**			These structures must be located in memory
**			which is accessible by both the LANCE and
**			the host processor.
**
**  PARAMETERS:		
**
**	Input:		retptr	: pointer to storage for return code
**
**	Output:
**
**	Returns:	retptr	: indicates the outcome of this
**				  function (i.e. holds return code)
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ma_int_mem(retptr)

register ULONG  *retptr; 	/* Pointer To Storage For Return Code */
{
	VOID	 ma_int_rng();			/* Initialize Ring */

	register struct ringptr  *rptr;		/* Ring Pointer */

	register struct txbufrng *brptr;	/* Buffer Ring Pointer */

	register BDB	*bptr;			/* BDB Pointer */

	register ULONG	memptr;			/* Temp Pointer */

/*
**  First, a memory block is obtained for the LANCE Init Block and
**  Ethernet Registration Table, from the Event Parameter Block (EPB)
**  buffer pool. The Registration Table starts after the Init Block.
*/

	if ((ma_lnc_iblkptr = (struct initblk *)o_getepb(EPB_PID)) != NULL)
	{
		ma_reg_tbl = (UBYTE *)ma_lnc_iblkptr + sizeof(struct initblk);

		o_clean(ma_lnc_iblkptr, sizeof(struct initblk));

		o_clean(ma_reg_tbl, (sizeof(MA_REG) * REG_TBL_SIZE));
	}
	else
	{
		*retptr = MA_ERR_MEM;

		LOG("MAC: Couldn't get buffer for LANCE Init Block");
		LOG("or Registration Table\n");

		return;
	}

/*
**  Next, if no errors have occurred, a memory block is obtained for
**  the LANCE Transmit Descriptor Ring from the Buffer Descriptor
**  Block (BDB) buffer pool.
**
**  If no errors have occurred, the returned address is converted to
**  quadword boundary and all Transmit Ring entries are cleared and
**  claimed by the processor. Also the address of the Transmit Desc
**  Ring is placed in the Init Block.
*/

	rptr = &ma_lnc_txrgptr;

	if ((memptr = (ULONG)o_getbdb(BDB_PID)) != NULL)
	{
		memptr += 7;
		memptr &= (-8);

		rptr->top = (struct ringentry *)memptr;

		ma_int_rng(&ma_lnc_iblkptr->txring, rptr, TXRLEN);
	}
	else
	{
		*retptr = MA_ERR_MEM;

		LOG("MAC: Couldn't get buffer for TX ring\n");

		return;
	}

/*
**  Next, if no errors have occurred, a memory block is obtained for
**  the LANCE Transmit Buffer Ring from the BDB buffer pool.
**
**  If no errors have occurred, Transmit Buffer Ring is cleared.
*/

	brptr = &ma_txrgptr;

	if ((memptr = (ULONG)o_getbdb(BDB_PID)) != NULL)
	{
		memptr += 7;
		memptr &= (-8);

		brptr->top = (ULONG *)memptr;

		brptr->bot  = brptr->top + ((1 << TXRLEN) - 1);
		brptr->tail = brptr->head = brptr->top;

		o_clean(brptr->top, (ULONG)(brptr->bot + 1) - (ULONG)(brptr->top));
	}
	else
	{
		*retptr = MA_ERR_MEM;

		LOG("MAC: Couldn't get buffer for TX buffer ring\n");

		return;
	}


/*
**  Next, if no errors have occurred, a memory block is obtained for
**  the LANCE Receive Descriptor Ring from the BDB pool.
**
**  If no errors have occurred, the returned address is converted to
**  quadword boundary and all Receive Ring entries are cleared and
**  claimed by the processor. Also the address of the Receive Desc
**  Ring is placed in the Init Block.
**
**  The function will then attempt to fill the LANCE Receive
**  Descriptor Ring with empty Buffer Descriptor Blocks for
**  storage of incoming data.
*/

	rptr = &ma_lnc_rxrgptr;

	if ((memptr = (ULONG)o_getbdb(BDB_PID)) != NULL)
	{
		memptr += 7;
		memptr &= (-8);

		rptr->top = (struct ringentry *)memptr;

		ma_int_rng(&ma_lnc_iblkptr->rxring, rptr, RXRLEN);

		while ((*retptr == RET_OK) && (rptr->tail->md[2] == 0))
			ma_rx_gbuf(retptr);
	}
	else
	{
		*retptr = MA_ERR_MEM;

		LOG("MAC: Couldn't get buffer for RX ring\n");
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ma_int_rng
**
**  DESCRIPTION:	This function initializes a LANCE descriptor
**			ring by loading the ring's Init Block field
**			and clearing eachof the ring's entries. It
**			is designed to initialize either the Transmit
**			or Receive Descriptor Ring. It assumes that
**			the ring's top and bottom pointers are
**			already set.
**
**			Upon entry to this function, the ring's top
**			pointer is assumed to be properly filled.
**
**  PARAMETERS:		
**
**	Input:		initptr		: pointer to ring's Init
**					  Block field
**
**			rptr		: pointer to ring's
**					  governing structure
**
**			rlen		: ring's length
**
**	Output:
**
**	Returns:
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ma_int_rng(initptr, rptr, rlen)

register ULONG	*initptr; 		/* Pointer To Ring's Init Block Field */
register struct ringptr *rptr;		/* Pointer To Ring's Governing Structure */
register UBYTE	rlen; 			/* Ring's Length */
{

/*
**  First, the LANCE Init Block field describing the ring is filled.
*/

	/*
	**  The higher word of the field contains the lower word of the
	**  ring's starting address. The lower byte of the lower word
	**  of the field holds the high byte of this (24-bit) starting
	**  address.
	*/

	*((USHORT *)initptr + 0)  = (USHORT)(rptr->top);

	*((USHORT *)initptr + 1)  = (USHORT)((ULONG)(rptr->top) >> 16);
	*((USHORT *)initptr + 1) &= (USHORT)0x00FF;
	
	/*
	**  Next, the ring's length is put into the top three bits
	**  of the lower word of the Init Block field.
	*/

	*((USHORT *)initptr + 1) |= (USHORT)(rlen << 13);

/*
**  Here the ring's head and tail pointers are initialized and all
**  ring entries are cleared and claimed by the processor.
*/

	rptr->bot  = rptr->top + ((1 << rlen) - 1);
	rptr->tail = rptr->head = rptr->top;

	o_clean(rptr->top, (ULONG)(rptr->bot + 1) - (ULONG)(rptr->top));
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ma_int_lnc
**
**  DESCRIPTION:	This function programs the LANCE registers to
**			execute chip initialization and to begin
**			interrupt-driven operation. The chip
**			initialization is controlled by polled
**			operation in order to simplify the LANCE
**			Interrupt Service Routine ma_isr().
**
**			Upon entry to this function, the Init Block
**			is assumed to be properly filled. Upon exit
**			from this function, the LANCE is operating
**			per the parameters of the Init Block with its
**			interrupts enabled and Register Address Port
**			pointing to CSR0.
**
**  PARAMETERS:		
**
**	Input:		ma_lnc_iblkptr		: LANCE Init Block
**						  pointer
**
**	Output:
**
**	Returns:
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ma_int_lnc()

{
	register DEFAULT i;			/* Counter */
	register USHORT  csr0;			/* CSR0 Contents */
	register BYTE	 *pcc_ptr;		/* MVME147 PCC LANCE Interrupt Control Address */
	register struct	 initblk *ma_iblkptr;	/* LANCE Initialization Block */

/*
**  First, the node's MAC address is moved into the LANCE
**  Init Block. Physical Address Bit (0) Zero is checked to
**  be zero. If it is not, an error is logged.
*/

	if ((ma_ts_addr[0] & 1) != 0)
	{
		LOG("MAC: Bad station address\n");
	}
	else
	{
		ma_iblkptr = ma_lnc_iblkptr;

		ma_iblkptr->padr[0] = ma_ts_addr[1];
		ma_iblkptr->padr[1] = ma_ts_addr[0];
		ma_iblkptr->padr[2] = ma_ts_addr[3];
		ma_iblkptr->padr[3] = ma_ts_addr[2];
		ma_iblkptr->padr[4] = ma_ts_addr[5];
		ma_iblkptr->padr[5] = ma_ts_addr[4];

	/*
	**  Next, the node's Group/Logical address mask is moved into
	**  the LANCE Initialization Block.
	*/

		ma_iblkptr->ladrf[0] = ma_ga_mask[3];
		ma_iblkptr->ladrf[1] = ma_ga_mask[2];
		ma_iblkptr->ladrf[2] = ma_ga_mask[1];
		ma_iblkptr->ladrf[3] = ma_ga_mask[0];

	/*
	**  Next, the Bus Master interface is defined by setting the
	**  BCON, ACON, and BSWP bits as necessary in CSR3.
	*/

		ma_set_rap(RAP_CSR3);

		ma_set_rdp(CSR3_BSWP | CSR3_BCON);

	/*
	**  Next, the location of the LANCE Init Block is loaded into 
	**  CSR1 and CSR2. The lower byte of CSR2 gets the high order
	**  byte of the (24-bit) base address of the Init Block. CSR1
	**  gets the low two bytes of this address.
	*/

		ma_set_rap(RAP_CSR2);

		ma_set_rdp((ULONG)ma_iblkptr >> 16);

		ma_set_rap(RAP_CSR1);

		ma_set_rdp(ma_iblkptr);

	/*
	**  Now the chip is ready to initialize itself, and the INIT bit
	**  in CSR0 is set to initiate this self-initialization. When
	**  the initializatin is complete, the chip sets the IDON bit in
	**  CSR0. This would generate an interrupt, except that the INEA
	**  bit was cleared above; therefore the IDON bit is polled to
	**  sense the completion of chip initialization. The IDON bit is
	**  then written to clear it for normal procesing.
	*/

		ma_set_rap(RAP_CSR0);

		ma_set_rdp(CSR0_INIT);

		for (i = 50000; i != 0; --i)
		{
			csr0 = ma_get_rdp();

			if ((csr0 & (USHORT)(CSR0_ERR | CSR0_IDON)) != 0)
				break;
		}

	/*
	**  Now that the chip is initialized, the STRT (start)
	**  and INEA (interrupt enable) bits are set in CSR0 to
	**  begin LANCE interrupt-driven operations if no initialization
	**  errors occured.
	*/

		if ((i != 0) && ((csr0 & CSR0_ERR) == 0))
		{
			ma_set_rdp(CSR0_IDON | CSR0_INEA | CSR0_STRT);

			*(pcc_ptr = (BYTE *)PCC_LANCE) = MVME147_LNC_ILEV | PCC_LANIEN;

			LOG("MAC: LANCE INEA and STRT, LANIEN = %x\n", *pcc_ptr);
		}
	}
}
