static char ident[] = " @(#)ma_isr.c	35.2        6/17/88	Copyright MOTOROLA, INC. ";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	ma_isr.c
**
**  DESCRIPTION:	This module will hold all the MAC interrupt
**			service routines.
**
**  CONTENTS:		ma_isr()
**			ma_isr_rx()
**			ma_isr_tx()
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:
**
**********************************************************************/

/*
**	Include Files
*/

#include <sys/osi/common/types.h>	/* Abstract Type Definitions */
#include <sys/osi/common/system.h>	/* System Function Definitions */
#include <sys/osi/common/bdb.h>		/* BDB Type Definitions */
#include <sys/osi/common/epb.h>		/* EPB Type Definitions */
#include <sys/osi/common/selsizes.h>	/* MAC Address Size */
#include <sys/osi/common/ma_epb.h>	/* MAC EPB Type Definitions */
#include <sys/osi/common/macros.h>	/* Macro Definitions */

#include "ma_lance.h"			/* LANCE Controller Definitions */
#include "ma_def.h"			/* MAC Node Definitions */
#include "ma_globals.h"			/* Global MAC Node Definitions */

/*
**	External Functions
*/

extern	VOID	o_clean();	/* Clear Bytes */
extern	VOID	o_retepb();	/* Return A EPB */
extern	VOID	o_retbdb();	/* Return A BDB */
extern	VOID	o_post();	/* Post A EPB */
extern	EPB	*o_getepb();	/* Get An EPB */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ma_isr
**
**  DESCRIPTION:	This function is the central processor of LANCE
**			interrupts. It takes care of cleanly entering
**			and exiting interrupt-level processing,
**			determinig the cause of the interrupt, and
**			invoking other MAC functions accordingly.
**
**  PARAMETERS:		
**
**	Input:		csr0	: control status register 0
**				  contents
**
**	Output:		The MAC Statistics Table is updated with
**			error and packet transmission/reception
**			information.
**
**	Returns:
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ma_isr(csr0)
register USHORT csr0;	 	/* Contents Of LANCE CSR0 */

{
	VOID	ma_isr_tx(),	/* Process Transmit Interrupt */
		ma_isr_rx();	/* Process Receive Interrupt */

/*
**  First, update the interrupt counter.
*/

	++ma_ir_ctr;

/*
**  The following section will check Control/Status Register
**  Zero (0) for hardware related errors and update the
**  statistics table accordingly.
*/

	/*
	**  The BABL bit is set when the transmitter has been on
	**  the channel longer than time requred to send the 
	**  maximum length packet.
	*/

	if ((csr0 & CSR0_BABL) != 0)
		++(ma_stat_tbl.hard_babl);

	/*
	**  The CERR bit indicates that the collision input to
	**  the chip failed to activate within 2us after a
	**  chip-initiated transmission was completed.
	*/

	if ((csr0 & CSR0_CERR) != 0)
		++(ma_stat_tbl.hard_cerr);

	/*
	**  The MISS bit is set when the receiver loses a packet
	**  because it does not own the next receive ring entry.
	**  It indicates the irretrievable loss of incoming
	**  data.
	*/

	if ((csr0 & CSR0_MISS) != 0)
		++(ma_stat_tbl.hard_miss);

	/*
	**  The MERR bit is set when the chip is in the Bus
	**  Master mode and has to wait too long for access
	**  to memory. When this error is detected, the LANCE
	**  turns off the receiver and transmitter, requiring
	**  a call to ma_rst() to restart the LANCE.
	*/

	if ((csr0 & CSR0_MERR) != 0)
		++(ma_stat_tbl.hard_merr);

/*
**  The following code processes all pending receive
**  interrupts. It calls ma_isr_rx(), which processes one
**  received packet, repeatedly until the next ring entry is
**  empty or the next ring entry is owned by the LANCE.
*/

	if ((csr0 & CSR0_RINT) != 0)
		ma_isr_rx();

/*
**  The following code processes all pending transmit
**  interrupts. It calls ma_isr_tx(), which processes one
**  transmitted packet, repeatedly until the next ring entry is
**  empty or the next ring entry is owned by the LANCE.
*/

	if ((csr0 & CSR0_TINT) != 0)
		ma_isr_tx();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ma_isr_rx
**
**  DESCRIPTION:	This function responds to a LANCE receive
**			interrupt, which signals the reception of a
**			packet from the LAN. It cleans up the LANCE
**			Receive Descriptor Ring and interprets the
**			MAC header, processing the packet further
**			as necessary according to its format and
**			content.
**
**  PARAMETERS:		
**
**	Input:		ma_lnc_rxrgptr	: LANCE Transmit Desc Ring
**
**
**	Output:		The MAC Statistics Table is updated with
**			error and packet reception information.
**
**	Returns:
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ma_isr_rx()

{
	register struct ringentry *rptr; /* Pointer To Ring Entry */

	register BDB	*bdbptr;  /* Pointer To BDB */

	register EPB	*epbptr;  /* Pointer To EPB */

	register UBYTE	*dataptr; /* Pointer To Start Of Data */

	register USHORT md_stat;  /* RX Status */
	register USHORT error;	  /* Local Error Code Variable */

/*
**  Determine if the head entry is a valid entry to process.
*/

	rptr = ma_lnc_rxrgptr.head;

	while (((rptr->md[3] != 0) || ((rptr->md[1] & RMD1_ERR) != 0)) && ((rptr->md[1] & RMD1_OWN) == 0))
	{

	/*
	**  First, we must initialize all function variables.
	*/

		error = RET_OK;

	/*
	**  The next section of this function pulls the newly-filled Data
	**  Block out of the Receive Descriptor Ring, stores the data
	**  pointer and data length parameters in the BDB, and updates
	**  the receive ring pointers.
	**
	**  Also the MAC sublayer statistics table is updated to reflect the
	**  reception of this packet.
	*/

		/*
		**  The latest ring entry to read is pointed to by the ring's 
		**  head pointer. The address of the buffer associated with
		**  this entry is extracted. This address points to the MAC
		**  header field of the packet, so the pointer to the BDB is
		**  calculated by subtracting an offset equal to the length
		**  of the BDB header from the address.
		*/

		dataptr = READ_RING(rptr);

		bdbptr = (BDB *)((ULONG)dataptr - BUF_FILL - BUF_HDR_SIZE - (ULONG)(sizeof(BDB) - PDU_SIZE));

		++ma_rx_ctr;

		/*
		**  Next, the buffer length is loaded into the BDB.
		*/

		bdbptr->buff_length = rptr->md[3];

		/*
		**  Determine if the packet was received with an framing
		**  error.
		*/

		if ((md_stat = rptr->md[1]) & RMD1_FRAM)
			++(ma_stat_tbl.rx_fram);

		/*
		**  Determine if the packet was received with an overflow
		**  error.
		*/

		if (md_stat & RMD1_OFLO)
			++(ma_stat_tbl.rx_oflo);

		/*
		**  Determine if the packet was received with an CRC
		**  error.
		*/

		if (md_stat & RMD1_CRC)
			++(ma_stat_tbl.rx_crc);

		/*
		**  Determine if the packet was received with an buffer
		**  error.
		*/

		if (md_stat & RMD1_BUFF)
			++(ma_stat_tbl.rx_buff);

		/*
		**  Then the function reads a summary of fatal error bits
		**  (framing, overflow, CRC or buffer errors) and sets the
		**  error code accordingly.
		*/

		if (md_stat & RMD1_ERR)
			error = MA_ERR_RXP;

	/*
	**  The ring entry must be cleared and the ring head pointer
	**  must be updated.
	*/

		o_clean(rptr, sizeof(struct ringentry));

		rptr = ma_lnc_rxrgptr.head = INCR_RING(rptr, ma_lnc_rxrgptr);

	/*
	**  The next section of this function will attempt to build a
	**  MA_DATA_INDICATION service request to pass back to the MAC sublayer.
	**  If an EPB cannot be obtained, all data buffers will be returned.
	*/

		if ((epbptr = o_getepb(EPB_PID)) != (EPB *)NULL)
		{
			epbptr->src_id   = MAC_QID;
			epbptr->dst_id   = MAC_QID;
			epbptr->epb_size = sizeof(EPB);
			epbptr->type     = MA_DATA_INDICATION;
			epbptr->status   = error;
			epbptr->bdb      = bdbptr;

			o_post(epbptr);
		}
		else
			o_retbdb(bdbptr);
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ma_isr_tx
**
**  DESCRIPTION:	This function responds to a LANCE transmit
**			completion interrupt by cleaning up the LANCE
**			Transmit Descriptor Ring and by notifying
**			the originator of the transmission of its
**			outcome.
**
**  PARAMETERS:		
**
**	Input:		ma_lnc_txrgptr	: LANCE Transmit Desc Ring
**
**
**	Output:		The MAC Statistics Table is updated with
**			error and packet transmission information.
**
**	Returns:
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ma_isr_tx()

{
	register struct ringentry *rptr;   /* Pointer To Ring Entry */

	register BDB	*bdbptr;  /* Pointer To Buffer Desc Block */

	register MA_EPB	*epbptr;  /* Pointer To MAC EPB */

	register UBYTE	*dataptr; /* Pointer To Top Of Data In Packet */

	register ULONG	*brgptr;  /* Buffer Ring Address */

	register USHORT	md_stat;  /* TX Status */

	register USHORT error;	  /* Local Error Code Variable */

/*
**  Determine if the head entry is a valid entry to process.
*/

	rptr   = ma_lnc_txrgptr.head;

	brgptr = ma_txrgptr.head;

	while ((rptr->md[2] != 0) && (((rptr->md[1] & TMD1_OWN) == 0) && ((rptr->md[1] & TMD1_STP) != 0)))
	{

	/*
	**  First, we must initialize all function variables.
	*/

		error = RET_OK;

	/*
	**  The first Data Block in the packet is extracted from the
	**  transmit ring's head entry. The pointer to the
	**  Buffer Desc Block is retrieved from its storage place directly
	**  above the transmitted data (see ma_data_req() processing). Also
	**  the pointer to the Event Parameter Block associated with the BDB
	**  is retrieved. This will be used to pass a MA_DATA.confirm to the
	**  originator with transmission status.
	*/

		epbptr = (MA_EPB *)(*brgptr);
	
		bdbptr = epbptr->hdr.bdb;

		++ma_tx_ctr;

	/*
	**  Next, the function reads the bits from the transmit ring's head
	**  entry (Transmit Message Descriptor One) which pertain to the MAC
	**  sublayer statistics and updates the statistics table as necessary.
	*/

		/*
		**  Determine if more than one but less than sixteen (16)
		**  retries were performed.
		*/

		if ((md_stat = rptr->md[1]) & TMD1_MORE)
			++(ma_stat_tbl.tx_more);

		/*
		**  Determine if exactly one retry was performed before
		**  the packet was transmitted.
		*/

		if (md_stat & TMD1_ONE)
			++(ma_stat_tbl.tx_one);

		/*
		**  Determine if transmission was deferred because the
		**  channel was busy.
		*/

		if (md_stat & TMD1_DEF)
			++(ma_stat_tbl.tx_def);

	/*
	**  Next, the function reads the bits from the transmit ring's head
	**  entry (Transmit Message Descriptor Three) which pertain to the MAC
	**  sublayer statistics and updates the statistics table as necessary.
	*/

		/*
		**  Determine if transmission was aborted due to
		**  buffer error.
		*/

		if ((md_stat = rptr->md[3]) & TMD3_BUFF)
			++(ma_stat_tbl.tx_buff);

		/*
		**  Determine if transmission was aborted due to
		**  underflow error.
		*/

		if (md_stat & TMD3_UFLO)
			++(ma_stat_tbl.tx_uflo);

		/*
		**  Determine if a late collision was detected.
		*/

		if (md_stat & TMD3_LCOL)
			++(ma_stat_tbl.tx_lcol);

		/*
		**  Determine if transmission was aborted due to loss
		**  of carrier.
		*/

		if (md_stat & TMD3_LCAR)
			++(ma_stat_tbl.tx_lcar);

		/*
		**  Determine if transmission was aborted due to excessive
		**  retries (sixteen).
		*/

		if (md_stat & TMD3_RTRY)
		{
			++(ma_stat_tbl.tx_rtry);

			ma_stat_tbl.tx_tdr = (ULONG)(rptr->md[3] & TMD3_TDR);
		}

	/*
	**  Finally, the function reads a summary of fatal error bits (late
	**  collision, loss of carrier, silo underflow, excessive retries)
	**  and sets the return code accordingly.
	*/

		if (rptr->md[1] & TMD1_ERR)
			error = MA_ERR_TXP;

	/*
	**  Next, the function clears the ring entries and updates the 
	**  Transmit ring pointers.
	*/

		o_clean(rptr, sizeof(struct ringentry));

		rptr = ma_lnc_txrgptr.head = INCR_RING(rptr, ma_lnc_txrgptr);

		*brgptr = 0;

		brgptr = ma_txrgptr.head = INCR_RING(brgptr, ma_txrgptr);

	/*
	**  Here the function generates an MA_DATA_CONFIRM to
	**  be sent to the MAC sublayer.
	*/
	
		epbptr->hdr.type   = MA_DATA_CONFIRM;
		epbptr->hdr.status = RET_OK;
		epbptr->hdr.bdb    = (BDB *)NULL;

		epbptr->status = error;

		o_retbdb(bdbptr);

		o_post(epbptr);
	}
}
