/*@(#)bm.h	35.1*/
/*
**	Buffer Management Window Structure (BM_WINDOW)
**
**	This structure is used to implement flow control windows via the
**	BPPIR buffer management routines.  An application can set its buffer
**	window size via a bm_set_window() function call.  Once set, any IP
**	using that window will not be allowed to acquire more than the number
**	of buffers specified by its "maximum".  If no limits are specified for
**	particular window ID, the additional overhead of keeping track of current
**	buffer counts by window is bypassed.  If multiple IP's are using the
**	same buffer pool, each one can individually set its own window.  If
**  a single application is registered as multiple IP's, it can use a 
**	single window per IP it wishes to have flow controlled or may use a
**	combined window for all of its IPs, or any combination.  Applications
**	on either side of a buffered pipe may use windows without forcing 
**	the other side to use them.
*/
typedef USHORT	BM_WID_T;

typedef struct bm_pend_t
{
	struct bm_pend_t	*next_pend;		/* list of blocked requests for this
									resource window */
	PTF			func_code;		/* either a function code STD_PEND or
									a pointer to a user function to call
									when the resource is available */
	ULONG		func_arg;		/* an argument to be used in calling the
									user function or to wake the suspended
									process. */
	BM_WID_T	wind_id;		/* resource window for which the request
									was made */
	ULONG		resource_id;	/* ID of the resource for which this
									request is pending */
} BM_PEND_T;

typedef struct bm_window
{
	USHORT	current;			/* Current count of buffers allocated
									to this interface point			*/
	USHORT	maximum;			/* Maximum number of buffers that can
									be allocated to this IP			*/
	BM_PEND_T	*next_pend;		/* pointer to head of list of requests
									blocked on this resource window */
	BM_PEND_T	*pend_tail;		/* pointer to tail of request list */
} BM_WINDOW;

/*
**  Pool Control Block (PCB)
**
**	The pool control block is used by BPPIR buffer management to control
**	the buffer pools created through create_msg_pool() requests.
*/

typedef struct pcb_t
{
	struct pcb_t	*next;		/* forward link to next pcb or zero 	*/
	BM_POOL_T		pool;		/* unique identifier of this pool 		*/
	USHORT			buf_size;	/* Size of buffers in this pool			*/
	BP_SYSHDR_T		*get_ptr;	/* pointer to first mpb in this pool 	*/
	BP_SYSHDR_T		*rel_ptr;	/* pointer to last mpb in dest cpu pool	*/
	BM_PEND_T		pend_list;	/* head of resource pend list, kept as a
									structure with the pointer in next_pend
									for coding efficiency in the management
									routines.							*/
	BM_PEND_T		*pend_tail;	/* tail of request pend list 			*/
	PTF				local_addr;	/* pointer to function to map VMEbus	*/
								/* address of mpb to local address space*/
	PTF				vme_addr;  	/* pointer to function to map local 	*/
								/* address of mpb to VME address space	*/
	BYTE			*start;		/* start of this buffer pool area 		*/
	BYTE			*end;		/* end of this buffer pool area			*/
	ULONG			pool_size;	/* size of this buffer pool area (bytes)*/		
	SHORT			orig_count;	/* original count of buffers created	*/
	SHORT			free_count;	/* current count of buffers in pool		*/
	BP_SYSHDR_T		*last_proc;	/* pointer to last mpb released by dest	*/
								/* that has been processed for window	*/
								/* control and available count			*/
	BP_SYSHDR_T		*last_alloc;/* pointer to list of buffers that have */
								/* have been sent to other CPU but		*/
								/* have not been returned or released.	*/
} BM_PCB_T;


/* Buffer management request message type */

typedef struct bfm_req_t
{
	BP_MPB_T	hdr;				/* MPB standard header */
	USHORT		type;				/* request type */
	USHORT		response_code;		/* status returned for request */
	BM_POOL_T	pool;				/* ID of pool for which request is made */
	USHORT		window_id;			/* ID of window used for request */
	USHORT		count;				/* number of mpb's requested or sent */
	BP_MPB_T	*mpb_link;			/* pointer to the mpb's transferred */
} BFM_REQ_T;

/* buffer pool create request and response message types */

typedef struct bm_crreq_t
{
	BP_MPB_T	hdr;				/* MPB standard header */
	USHORT		type;				/* request type */
	BM_POOL_T	pool;				/* ID of pool */
	USHORT		resp_code;			/* status returned with response */
	BYTE		*addr;				/* address of block used for pool */
	ULONG		mem_size;			/* size in bytes of block for pool */
} BM_CRREQ_T;

typedef struct bm_inreq_t
{
	BP_MPB_T	hdr;				/* MPB standard header */
	USHORT		type;				/* request type */
	USHORT		resp_code;			/* status returned with response */
	BM_POOL_T	creator_pool_id;	/* ID of pool */
	BM_POOL_T	manager_pool_id;	/* ID of pool */
	ULONG		buf_size;			/* size of buffers in pool */
	USHORT		count;				/* number of buffers for sub-pool */
	BP_MPB_T	*get_buf;			/* first buffer in local sub-pool */
	BP_MPB_T	*ret_buf;			/* last buffer in local sub-pool */
	BP_MPB_T	*first_buf;			/* first buffer created for local sub-pool */
	BP_MPB_T	*last_buf;			/* last buffer created for local sub-pool */
} BM_INREQ_T;

/* Message type codes for buffer management */

#define	BM_CREATE_POOL_REQ		0x01	/* Create a buffer pool */
#define	BM_INIT_POOL_REQ		0x02	/* Initialize a buffer pool */
#define	BM_FREE_POOL_REQ		0x03	/* Free a buffer pool */
#define	BM_GET_BUF_REQ			0x04	/* Get buffer request */
#define	BM_RET_BUF_REQ			0x05	/* Return buffer request */
#define BP_AVAIL_BUF			0x07	/* Request current pool count */

#define	BM_CREATE_POOL_RSP		0x11	/* Response to create pool request */
#define BM_INIT_POOL_RSP		0x12	/* Response to init pool request */
#define BM_FREE_POOL_RSP		0x13	/* Response to free pool request */
#define BM_GET_BUF_RSP			0x14	/* response to get buf request */
#define BM_RET_BUF_RSP			0x15	/* Response to return buf request */

/* Error codes for buffer management */
/* Function return codes */

#define	BM_SUCCESS	BPP_SUCCESSFUL			/* successful completion */
#define	BM_BAD_POOL_ID		0xf100			/* invalid pool id parameter */
#define BM_CREATE_PARM		0xf101			/* bad argument to create call */
#define	BM_INSUFF_MEMORY	0xf102			/* uable to get shared memory */

#define BM_NO_WINDOW_AVAIL	0xf110			/* no window structure available */
#define BM_WINDOWS_PARM		0xf111			/* bad parameter to windows
												management parameter */

/* bp_get_mpb() and bp_ret_mpb() return codes and definitions */

/* request func_code argument definitions */
#define	NO_PEND		0x00000000				/* no blocking on availability */
#define	STD_PEND	0xffffffff				/* use standard task blocking */

/* get error return definitions */
#define	BM_NO_MPB_AVAIL		0xf101			/* no buffers -- used with NO_PEND */
#define	BM_WINDOW_FULL		0xf102			/* window is full */
#define	BM_INVALID_WINDOW	0xf103			/* invalid window id */

/* ret call errors */
#define	BM_INVALID_MPB		0xf110			/* bad MPB or no pipe to owner */

/* resp_code field definitions */
